using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.Unicode;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;
using Microsoft.Win32.SafeHandles;
using Newtonsoft.Json;

namespace FitConnect.Encryption;

public class CertificateHelper {
    private readonly ILogger? _logger;

    public CertificateHelper(ILogger? logger = null) {
        _logger = logger;
    }

    internal bool ValidateCertificate(string keyJson, LogLevel logLevel,
        X509Certificate2[]? rootCertificate = null) =>
        ValidateCertificate(new JsonWebKey(keyJson), logLevel, rootCertificate);

    internal bool ValidateCertificate(X509Certificate2 certificate,
        out X509ChainStatus[] chainStatus,
        X509Certificate2[]? rootCertificate = null,
        X509Certificate2[]? extras = null,
        LogLevel logLevel = LogLevel.Warning) {
        // Working notes
        // https://git.fitko.de/fit-connect/planning/-/issues/142
        // https://docs.microsoft.com/en-us/dotnet/api/system.security.cryptography.x509certificates.x509chain.build?view=net-5.0

        var certificateChain = new X509Chain();
        certificateChain.ChainPolicy.ExtraStore.AddRange(extras ?? Array.Empty<X509Certificate2>());

        _logger?.LogDebug("Issuers: {Issuer}", certificate.Issuer);

        if (rootCertificate != null) {
            certificateChain.ChainPolicy.TrustMode = X509ChainTrustMode.CustomRootTrust;
            certificateChain.ChainPolicy.CustomTrustStore.AddRange(rootCertificate);
            certificateChain.ChainPolicy.RevocationMode = X509RevocationMode.Online;
            certificateChain.ChainPolicy.RevocationFlag = X509RevocationFlag.EndCertificateOnly;
            certificateChain.ChainPolicy.VerificationFlags = X509VerificationFlags.NoFlag;
            _logger?.LogDebug("Using custom root certificate");
        }
        else {
            certificateChain.ChainPolicy.RevocationMode = X509RevocationMode.NoCheck;
            certificateChain.ChainPolicy.RevocationFlag = X509RevocationFlag.EntireChain;
            certificateChain.ChainPolicy.VerificationFlags = X509VerificationFlags.AllFlags;
            certificateChain.ChainPolicy.DisableCertificateDownloads = false;
        }


        var result = certificateChain.Build(certificate);

        chainStatus = certificateChain.ChainStatus
            .Where(s => s.Status != X509ChainStatusFlags.PartialChain).ToArray();

        var statusAggregation = certificateChain.ChainStatus.Aggregate("",
            (r, s) => r + "\n\t - " + s.Status + ": " + s.StatusInformation);

        if (certificateChain.ChainStatus.Length > 0)
            _logger?.Log(logLevel, "Certificate status: {ObjStatusInformation}",
                statusAggregation);

        return result;
    }

    internal bool ValidateCertificate(JsonWebKey key, LogLevel logLevel = LogLevel.Error,
        X509Certificate2[]? root = null) {
        var certificates = key.X5c.Select(s => new X509Certificate2(Convert.FromBase64String(s)))
            .ToList();

        var fitConnectRequirements = key.MatchesFitConnectRequirements();
        if (!fitConnectRequirements) {
            _logger?.Log(logLevel, "Certificate does not match FIT-Connect requirements");
            return false;
        }
        
        var valid = ValidateCertificate(certificates.First(), out _, root, certificates.ToArray(),
            logLevel);
        return valid && fitConnectRequirements;
    }
}

public static class CertificateExtensions {
    public static void ExportToPem(this X509Certificate2 certificate, string fileName) {
        StringBuilder builder = new StringBuilder();
        builder.AppendLine("-----BEGIN CERTIFICATE-----");
        builder.AppendLine(
            Convert.ToBase64String(certificate.RawData, Base64FormattingOptions.InsertLineBreaks));
        builder.AppendLine("-----END CERTIFICATE-----");
        var content = builder.ToString();

        File.WriteAllText(fileName + ".pem", content);
        File.WriteAllText(fileName + ".json", JsonConvert.SerializeObject(certificate));
    }

    public static bool MatchesFitConnectRequirements(this JsonWebKey key) {
        return key.X5c.Count == 3
            && key.KeySize == 4096
            && (key.KeyOps.Contains("wrapKey"));
    }
}
